package sdkms

import (
	"encoding/base64"
	"log"
	"os"
	"reflect"
	"strconv"
	"testing"
	"time"

	"golang.org/x/net/context"
)

const PurposeSetup string = "setup"
const PurposeTransiet string = "transient"
const PurposeAppLogin string = "applogin"

var RSA, AES, DES, DES3, EC, HMAC, SECRET ObjectType
var ECB, CBC, CBCNOPAD CryptMode
var NistP256 EllipticCurve
var DiffieHellman AgreeKeyMechanism
var SHA1, SHA256 DigestAlgorithm

var TestConfiguration struct {
	APIBasePath string `json:"apiBasePath"`
	Username    string `json:"userName"`
	Password    string `json:"passWord"`
	APIKey      string `json:"apiKey"`
}

var TransientValues struct {
	AccessToken string
	RSAKeyID    string
	AESKeyID    string
	DESKeyID    string
	DES3KeyID   string
	ECKeyID     string
	WrappedKey  string
}

var TestRequisites struct {
	AccessToken string
	AccountID   string
	GroupID     string
	AppID       string
	AppToken    string
	RSAKeyID    string
	AESKeyID    string
	DESKeyID    string
	DES3KeyID   string
	AgreeKey1   SobjectDescriptor
	AgreeKey2   SobjectDescriptor
	HMACKeyID   string
	Plain1      string
	Plain2      string
	Cipher1     string //Generated using RSA and Plain1
	Cipher2     string //Generated using AES and Plain2
	RSACipher   string
	AESCipher   string
	DESCipher   string
	DES3Cipher  string
	IV8Len      string
	IV16Len     string
	Hash1       string
	Hash2       string
	Sign1       string
	Sign2       string
}

var cfg = NewConfiguration()

//Client - API client to be used for testing
var Client = NewAPIClient(cfg)

/*TestMain - The entry and exit point to run tests
 */
func TestMain(m *testing.M) {
	setup()
	log.Printf("Executing Test cases...")
	code := m.Run()
	shutdown()
	os.Exit(code)
}

func setup() {
	log.Printf("Pre-tests setup...")
	readTestConfiguartions()
	initValues()
	login(PurposeSetup, TestConfiguration.Username, TestConfiguration.Password)
	createAccount()
	selectAccount()
	createGroup()
	createApp()
	createSecretAndAPIKey()
	createSecurityObjetcs()
	generateCiphers()
}

func shutdown() {
	log.Printf("Post-tests cleanup...")
	deleteSecurityObjetcs()
	deleteApp()
	deleteGroup()
	deleteAccount()
}

func readTestConfiguartions() {
	TestConfiguration.APIBasePath = os.Getenv("FORTANIX_API_ENDPOINT")
	TestConfiguration.Username = os.Getenv("FORTANIX_USERNAME")
	TestConfiguration.Password = os.Getenv("FORTANIX_PASSWORD")
	TestConfiguration.APIKey = os.Getenv("FORTANIX_API_KEY")

	if TestConfiguration.APIKey == "" {
		log.Fatal("API Key not provided. Add env variable FORTANIX_API_KEY")
	}
	if TestConfiguration.APIBasePath == "" {
		log.Fatal("API Endpoint not provided. Add env variable FORTANIX_API_ENDPOINT")
	}
	cfg.BasePath = TestConfiguration.APIBasePath
	if TestConfiguration.Username == "" {
		log.Fatal("Test Username not provided. Add env variable FORTANIX_USERNAME")
	}
	if TestConfiguration.Password == "" {
		log.Fatal("Test User Password not provided. Add env variable FORTANIX_PASSWORD")
	}
}

func initValues() {

	RSA = ObjectType_RSA
	AES = ObjectType_AES
	DES = ObjectType_DES
	DES3 = ObjectType_DES3
	EC = ObjectType_EC
	HMAC = ObjectType_HMAC
	SECRET = ObjectType_SECRET

	ECB = CryptMode_ECB
	CBC = CryptMode_CBC
	CBCNOPAD = CryptMode_CBCNOPAD

	NistP256 = EllipticCurve_NIST_P256

	DiffieHellman = AgreeKeyMechanism_HELLMAN

	SHA1 = DigestAlgorithm_SHA1
	SHA256 = DigestAlgorithm_SHA256

	TestRequisites.Plain1 = "ABCDEFGHIJKLMNOP"
	TestRequisites.Plain2 = "PONMLKJIHGFEDCBA"
	TestRequisites.IV8Len = "00000000"
	TestRequisites.IV16Len = "0000000000000000"
	TestRequisites.Hash1 = "ZgxCVhOwakpb0K69Jb/MGtqudxk="
	TestRequisites.Hash2 = "ZheqiKcua1JriMvO2jiKe1Kg6FYUihLZuEKc0qU6PqQ="
}

/*IsMatching - Match the fields of api response with expected response
@param want struct object expected from API response
@param got struct object returned in API response
@param checkFields []string array of fields to match or check non null based on flag matchValues
@param matchValues bool specify whether to mathc values or not. If false then will only check the value is not null
*/
func IsMatching(want interface{}, got interface{}, checkFields []string, matchValues bool) bool {

	matching := true
	g := reflect.ValueOf(got)
	if g.Kind() == reflect.Slice { //This means the response is array and we need to check each object of array
		w := reflect.ValueOf(want)

		if matchValues && g.Len() != w.Len() {
			return false
		}
		for i := 0; i < g.Len(); i++ {
			var wantValues interface{}
			if i < w.Len() {
				wantValues = w.Index(i).Interface()
			}
			matching = compareValues(wantValues, g.Index(i).Interface(), checkFields, matchValues)
			if !matching {
				break
			}
		}

	} else {
		matching = compareValues(want, got, checkFields, matchValues)
	}
	return matching

}

func compareValues(want interface{}, got interface{}, checkFields []string, matchValues bool) bool {
	matching := true

	for _, element := range checkFields {
		r1 := reflect.ValueOf(got)
		f1 := reflect.Indirect(r1).FieldByName(element)
		//log.Println("Field comparing - ", element, r1, f1)

		if matchValues { //compare values
			r2 := reflect.ValueOf(want)
			f2 := reflect.Indirect(r2).FieldByName(element)
			matching = reflect.DeepEqual(convertToProperType(f1), convertToProperType(f2))
		} else { //check only that response json has the field
			if reflect.DeepEqual(convertToProperType(f1), nil) {
				matching = false
			}
		}
		if !matching {
			log.Println("Field not matching/absent - ", element)
			break
		}
	}
	return matching
}

func convertToProperType(input reflect.Value) interface{} {

	switch k := input.Kind(); k {
	case reflect.String:
		return input.String()
	case reflect.Int:
		return input.Int()
	case reflect.Int32:
		return input.Interface().(int32)
	case reflect.Int64:
		return input.Interface().(int64)
	case reflect.Float32:
		return input.Interface().(float32)
	case reflect.Float64:
		return input.Interface().(float64)
	case reflect.Bool:
		return input.Interface().(bool)
	case reflect.Ptr:
		//log.Println("Found Pointer ", k, input.Elem())
		return input.Elem()
	default:
		log.Println("unexpected type - ", k)
		return nil
	}

}

/*
GetContext - Get the context for a test
@param test name of the test
*/
func GetContext(t *testing.T) context.Context {
	authContext := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	appContext := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AppToken)
	switch t.Name() {
	case "TestAuthenticationApiService_Authorize":
		return context.WithValue(context.Background(), ContextBasicAuth, BasicAuth{
			UserName: TestConfiguration.Username,
			Password: TestConfiguration.Password,
		})
	case "TestAuthenticationApiService_Terminate":
		return context.WithValue(context.Background(), ContextAccessToken, TransientValues.AccessToken)
	case "TestEncryptionAndDecryptionApiService_Decrypt", "TestEncryptionAndDecryptionApiService_DecryptEx", "TestEncryptionAndDecryptionApiService_DecryptFinal",
		"TestEncryptionAndDecryptionApiService_DecryptFinalEx", "TestEncryptionAndDecryptionApiService_DecryptInit", "TestEncryptionAndDecryptionApiService_DecryptInitEx",
		"TestEncryptionAndDecryptionApiService_DecryptUpdate", "TestEncryptionAndDecryptionApiService_DecryptUpdateEx", "TestEncryptionAndDecryptionApiService_Encrypt",
		"TestEncryptionAndDecryptionApiService_EncryptEx", "TestEncryptionAndDecryptionApiService_EncryptFinal", "TestEncryptionAndDecryptionApiService_BatchEncrypt",
		"TestEncryptionAndDecryptionApiService_BatchDecrypt", "TestEncryptionAndDecryptionApiService_EncryptFinalEx", "TestEncryptionAndDecryptionApiService_EncryptInit",
		"TestEncryptionAndDecryptionApiService_EncryptInitEx", "TestEncryptionAndDecryptionApiService_EncryptUpdate", "TestEncryptionAndDecryptionApiService_EncryptUpdateEx",
		"TestSecurityObjectsApiService_AgreeKey", "TestSecurityObjectsApiService_DeriveKey", "TestSecurityObjectsApiService_GetSecurityObject",
		"TestSecurityObjectsApiService_GenerateSecurityObject", "TestSecurityObjectsApiService_DeriveKeyEx", "TestSecurityObjectsApiService_PersistSecurityObject",
		"TestSecurityObjectsApiService_ImportSecurityObject", "TestSecurityObjectsApiService_GetSecurityObjects", "TestSecurityObjectsApiService_GetSecurityObjectValueEx",
		"TestSignAndVerifyApiService_BatchSign", "TestSignAndVerifyApiService_BatchVerify", "TestSignAndVerifyApiService_Sign", "TestSignAndVerifyApiService_SignEx",
		"TestSignAndVerifyApiService_Verify", "TestSignAndVerifyApiService_VerifyEx", "TestWrappingAndUnwrappingApiService_WrapKey", "TestWrappingAndUnwrappingApiService_WrapKeyEx",
		"TestWrappingAndUnwrappingApiService_UnwrapKey", "TestWrappingAndUnwrappingApiService_UnwrapKeyEx":
		return appContext
	default:
		return authContext
	}

}

func setupTestCase(t *testing.T) func(t *testing.T) {
	switch t.Name() {
	case "TestAuthenticationApiService_Terminate":
		if TransientValues.AccessToken == "" {
			login(PurposeTransiet, TestConfiguration.Username, TestConfiguration.Password)
		}
	case "TestSecurityObjectsApiService_DeletePrivateKey", "TestSecurityObjectsApiService_DeleteSecurityObject":
		ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
		a := &SecurityObjectsApiService{
			client: Client,
		}
		TransientValues.RSAKeyID = createRSAKey(ctx, a)
	case "TestWrappingAndUnwrappingApiService_UnwrapKey", "TestWrappingAndUnwrappingApiService_UnwrapKeyEx":
		if TransientValues.WrappedKey == "" {
			ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AppToken)
			a := &WrappingAndUnwrappingApiService{
				client: Client,
			}
			TransientValues.WrappedKey = WrapKey(ctx, a)
		}
	}
	return func(t *testing.T) {
		//TODO: Add post test case execution for a test case if any
	}
}

func setupSubTest(t *testing.T) func(t *testing.T) {
	t.Log("setup sub test")
	return func(t *testing.T) {
		t.Log("teardown sub test")
	}
}

func login(purpose string, username string, password string) {

	ctx := context.WithValue(context.Background(), ContextBasicAuth, BasicAuth{
		UserName: username,
		Password: password,
	})

	a := &AuthenticationApiService{
		client: Client,
	}
	got, got1, _ := a.Authorize(ctx)

	if got.AccessToken != "" {
		if purpose == PurposeSetup {
			TestRequisites.AccessToken = got.AccessToken
		} else if purpose == PurposeTransiet {
			TransientValues.AccessToken = got.AccessToken
		} else if purpose == PurposeAppLogin {
			TestRequisites.AppToken = got.AccessToken
		}
	} else {
		log.Println("Error occured in pre-setup process - login(). \ngot = ", got, "\ngot1 = ", got1, "\nStopping tests execution")
		if purpose == PurposeSetup {
			shutdown()
			os.Exit(1)
		}
	}

}

func createAccount() {

	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &AccountsApiService{
		client: Client,
	}

	request := AccountRequest{Name: strconv.FormatInt(time.Now().UTC().UnixNano(), 10)}

	got, got1, _ := a.CreateAccount(ctx, request)

	if got.AcctId != "" {
		TestRequisites.AccountID = got.AcctId
	} else {
		log.Println("Error occurred in pre-setup process - createAccount(). \ngot = ", got, "\ngot1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}

}

func selectAccount() {

	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &AuthenticationApiService{
		client: Client,
	}

	request := SelectAccountRequest{AcctId: TestRequisites.AccountID}

	got, got1, _ := a.SelectAccount(ctx, request)

	if got1.StatusCode != 200 {
		log.Println("Error occurred in pre-setup process - selectAccount(). \ngot = ", got, "\ngot1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}

}

func createGroup() {
	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &GroupsApiService{
		client: Client,
	}

	request := GroupRequest{Name: strconv.FormatInt(time.Now().UTC().UnixNano(), 10), AcctId: TestRequisites.AccountID}

	got, got1, _ := a.CreateGroup(ctx, request)

	if got.GroupId != "" {
		TestRequisites.GroupID = got.GroupId
	} else {
		log.Println("Error occured in pre-setup process - createGroup(). \ngot = ", got, "\ngot1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
}

func createApp() {
	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &AppsApiService{
		client: Client,
	}
	
	appPermissions := []AppPermissions{
			AppPermissions_ENCRYPT,
			AppPermissions_DECRYPT,
			AppPermissions_WRAPKEY,
			AppPermissions_UNWRAPKEY,
			AppPermissions_DERIVEKEY,
			AppPermissions_MACGENERATE,
			AppPermissions_MACVERIFY,
			AppPermissions_MANAGE,
			AppPermissions_SIGN,
			AppPermissions_VERIFY,
			AppPermissions_AGREEKEY,
			AppPermissions_EXPORT}
	appGroups := make(map[string][]AppPermissions)	
	appGroups[TestRequisites.GroupID] = appPermissions

	request := AppRequest{
		Name:         strconv.FormatInt(time.Now().UTC().UnixNano(), 10),
		AddGroups:    appGroups,
		DefaultGroup: TestRequisites.GroupID,
	}

	got, got1, _ := a.CreateApp(ctx, request)

	if got.AppId != "" {
		TestRequisites.AppID = got.AppId
	} else {
		log.Println("Error occured in pre-setup process - createApp(). \ngot = ", got, "\ngot1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
}

func createSecretAndAPIKey() {
	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &AppsApiService{
		client: Client,
	}

	got, got1, _ := a.GetCredential(ctx, TestRequisites.AppID)

	if got.Credential != nil && got.Credential.Secret != "" {
		login(PurposeAppLogin, TestRequisites.AppID, got.Credential.Secret)
	} else {
		log.Println("Error occured in pre-setup process - createSecretAndAPIKey() \ngot = ", got, "\ngot1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}

}

func createSecurityObjetcs() {

	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &SecurityObjectsApiService{
		client: Client,
	}

	/*--------Generating RSA Key--------*/
	TestRequisites.RSAKeyID = createRSAKey(ctx, a)

	/*--------Generating AES Key--------*/
	TestRequisites.AESKeyID = createAESKey(ctx, a)

	/*--------Generating DES Key--------*/
	TestRequisites.DESKeyID = createDESKey(ctx, a)

	/*--------Generating DES3 Key--------*/
	TestRequisites.DES3KeyID = createDES3Key(ctx, a)

	/*--------Generating Agree Keys--------*/
	TestRequisites.AgreeKey1 = SobjectDescriptor{Kid: createAgreeKey(ctx, a)}
	TestRequisites.AgreeKey2 = SobjectDescriptor{Kid: createAgreeKey(ctx, a)}

	/*--------Generating HMAC Key--------*/
	TestRequisites.HMACKeyID = createHMACKey(ctx, a)

}

func generateCiphers() {
	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AppToken)
	a := &EncryptionAndDecryptionApiService{
		client: Client,
	}

	/*--------Generating RSA Cipher--------*/
	TestRequisites.RSACipher = RSAEncrypt(ctx, a, TestRequisites.Plain1, TestRequisites.RSAKeyID)
	TestRequisites.Cipher1 = TestRequisites.RSACipher

	/*--------Generating AES Cipher--------*/
	TestRequisites.AESCipher = AESEncrypt(ctx, a, TestRequisites.Plain1, TestRequisites.AESKeyID)
	TestRequisites.Cipher2 = AESEncrypt(ctx, a, TestRequisites.Plain2, TestRequisites.AESKeyID)

	/*--------Generating DES Cipher--------*/
	TestRequisites.DESCipher = DESEncrypt(ctx, a, TestRequisites.Plain1, TestRequisites.DESKeyID)

	/*--------Generating DES3 Cipher--------*/
	TestRequisites.DES3Cipher = DES3Encrypt(ctx, a, TestRequisites.Plain1, TestRequisites.DES3KeyID)

}

func createRSAKey(ctx context.Context, a *SecurityObjectsApiService) string {
	request := SobjectRequest{
		Name:    strconv.FormatInt(time.Now().UTC().UnixNano(), 10),
		ObjType: &RSA,
		GroupId: TestRequisites.GroupID,
		KeySize: 2048,
		KeyOps:  []KeyOperations{KeyOperations_EXPORT, KeyOperations_WRAPKEY, KeyOperations_UNWRAPKEY, KeyOperations_VERIFY, KeyOperations_SIGN, KeyOperations_ENCRYPT, KeyOperations_DECRYPT, KeyOperations_APPMANAGEABLE},
	}
	got, got1, _ := a.GenerateSecurityObject(ctx, request)

	if got.Kid != "" {
		return got.Kid
	} else {
		log.Println("Error occured in pre-setup process - createSecurityObjetcs() RSA. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func createAESKey(ctx context.Context, a *SecurityObjectsApiService) string {
	request := SobjectRequest{
		Name:    strconv.FormatInt(time.Now().UTC().UnixNano(), 10),
		ObjType: &AES,
		GroupId: TestRequisites.GroupID,
		KeySize: 128,
		KeyOps:  []KeyOperations{KeyOperations_EXPORT, KeyOperations_WRAPKEY, KeyOperations_UNWRAPKEY, KeyOperations_DERIVEKEY, KeyOperations_ENCRYPT, KeyOperations_DECRYPT, KeyOperations_APPMANAGEABLE},
	}

	got, got1, _ := a.GenerateSecurityObject(ctx, request)

	if got.Kid != "" {
		return got.Kid
	} else {
		log.Println("Error occured in pre-setup process - createSecurityObjetcs() AES. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func createDESKey(ctx context.Context, a *SecurityObjectsApiService) string {
	request := SobjectRequest{
		Name:    strconv.FormatInt(time.Now().UTC().UnixNano(), 10),
		ObjType: &DES,
		GroupId: TestRequisites.GroupID,
		KeySize: 56,
		KeyOps:  []KeyOperations{KeyOperations_EXPORT, KeyOperations_WRAPKEY, KeyOperations_UNWRAPKEY, KeyOperations_DERIVEKEY, KeyOperations_ENCRYPT, KeyOperations_DECRYPT, KeyOperations_APPMANAGEABLE},
	}

	got, got1, _ := a.GenerateSecurityObject(ctx, request)

	if got.Kid != "" {
		return got.Kid
	} else {
		log.Println("Error occured in pre-setup process - createSecurityObjetcs() DES. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func createDES3Key(ctx context.Context, a *SecurityObjectsApiService) string {
	request := SobjectRequest{
		Name:    strconv.FormatInt(time.Now().UTC().UnixNano(), 10),
		ObjType: &DES3,
		GroupId: TestRequisites.GroupID,
		KeySize: 168,
		KeyOps:  []KeyOperations{KeyOperations_EXPORT, KeyOperations_WRAPKEY, KeyOperations_UNWRAPKEY, KeyOperations_DERIVEKEY, KeyOperations_ENCRYPT, KeyOperations_DECRYPT, KeyOperations_APPMANAGEABLE},
	}

	got, got1, _ := a.GenerateSecurityObject(ctx, request)

	if got.Kid != "" {
		return got.Kid
	} else {
		log.Println("Error occured in pre-setup process - createSecurityObjetcs() DES3. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func createAgreeKey(ctx context.Context, a *SecurityObjectsApiService) string {
	request := SobjectRequest{
		Name:          strconv.FormatInt(time.Now().UTC().UnixNano(), 10),
		ObjType:       &EC,
		GroupId:       TestRequisites.GroupID,
		EllipticCurve: &NistP256,
		KeyOps:        []KeyOperations{KeyOperations_EXPORT, KeyOperations_VERIFY, KeyOperations_SIGN, KeyOperations_AGREEKEY, KeyOperations_APPMANAGEABLE},
	}

	got, got1, _ := a.GenerateSecurityObject(ctx, request)

	if got.Kid != "" {
		return got.Kid
	} else {
		log.Println("Error occured in pre-setup process - createSecurityObjetcs() Agree key 1. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func createHMACKey(ctx context.Context, a *SecurityObjectsApiService) string {
	request := SobjectRequest{
		Name:    strconv.FormatInt(time.Now().UTC().UnixNano(), 10),
		ObjType: &HMAC,
		GroupId: TestRequisites.GroupID,
		KeySize: 256,
		KeyOps:  []KeyOperations{KeyOperations_EXPORT, KeyOperations_MACVERIFY, KeyOperations_MACGENERATE, KeyOperations_APPMANAGEABLE},
	}

	got, got1, _ := a.GenerateSecurityObject(ctx, request)

	if got.Kid != "" {
		return got.Kid
	} else {
		log.Println("Error occured in pre-setup process - createSecurityObjetcs() HMAC key. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func RSAEncrypt(ctx context.Context, a *EncryptionAndDecryptionApiService, plain string, keyID string) string {
	request := EncryptRequest{Alg: &RSA, Plain: plain, Iv: base64.StdEncoding.EncodeToString([]byte(TestRequisites.IV16Len))}

	got, got1, _ := a.Encrypt(ctx, keyID, request)

	if got.Cipher != "" {
		return got.Cipher
	} else {
		log.Println("Error occured in pre-setup process - generateCiphers() using RSA key. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func AESEncrypt(ctx context.Context, a *EncryptionAndDecryptionApiService, plain string, keyID string) string {
	request := EncryptRequest{Alg: &AES, Mode: &CBC, Plain: plain, Iv: base64.StdEncoding.EncodeToString([]byte(TestRequisites.IV16Len))}

	got, got1, _ := a.Encrypt(ctx, keyID, request)

	if got.Cipher != "" {
		return got.Cipher
	} else {
		log.Println("Error occured in pre-setup process - generateCiphers() using AES key. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func DESEncrypt(ctx context.Context, a *EncryptionAndDecryptionApiService, plain string, keyID string) string {
	request := EncryptRequest{Alg: &DES, Mode: &CBC, Plain: plain, Iv: base64.StdEncoding.EncodeToString([]byte(TestRequisites.IV8Len))}

	got, got1, _ := a.Encrypt(ctx, keyID, request)

	if got.Cipher != "" {
		return got.Cipher
	} else {
		log.Println("Error occured in pre-setup process - generateCiphers() using DES key. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func DES3Encrypt(ctx context.Context, a *EncryptionAndDecryptionApiService, plain string, keyID string) string {
	request := EncryptRequest{Alg: &DES3, Mode: &CBC, Plain: plain, Iv: base64.StdEncoding.EncodeToString([]byte(TestRequisites.IV8Len))}

	got, got1, _ := a.Encrypt(ctx, keyID, request)

	if got.Cipher != "" {
		return got.Cipher
	} else {
		log.Println("Error occured in pre-setup process - generateCiphers() using DES3 key. \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func WrapKey(ctx context.Context, a *WrappingAndUnwrappingApiService) string {
	request := WrapKeyRequest{Alg: &AES, Mode: &CBC, Kid: TestRequisites.RSAKeyID, Iv: base64.StdEncoding.EncodeToString([]byte(TestRequisites.IV16Len))}
	got, got1, _ := a.WrapKey(ctx, TestRequisites.AESKeyID, request)

	if got.WrappedKey != "" {
		return got.WrappedKey
	} else {
		log.Println("Error occured in pre-setup process - WrapKey(). \ngot = ", got, "got1 = ", got1, "\nStopping tests execution")
		shutdown()
		os.Exit(1)
	}
	return ""
}

func deleteSecurityObjetcs() {

	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &SecurityObjectsApiService{
		client: Client,
	}

	a.DeleteSecurityObject(ctx, TestRequisites.RSAKeyID)
	a.DeleteSecurityObject(ctx, TestRequisites.AESKeyID)
	a.DeleteSecurityObject(ctx, TestRequisites.DESKeyID)
	a.DeleteSecurityObject(ctx, TestRequisites.DES3KeyID)
	a.DeleteSecurityObject(ctx, TestRequisites.AgreeKey1.Kid)
	a.DeleteSecurityObject(ctx, TestRequisites.AgreeKey2.Kid)
	a.DeleteSecurityObject(ctx, TestRequisites.HMACKeyID)

}

func deleteApp() {
	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &AppsApiService{
		client: Client,
	}

	a.DeleteApp(ctx, TestRequisites.AppID)
}

func deleteGroup() {
	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &GroupsApiService{
		client: Client,
	}
	a.DeleteGroup(ctx, TestRequisites.GroupID)
}

func deleteAccount() {
	ctx := context.WithValue(context.Background(), ContextAccessToken, TestRequisites.AccessToken)
	a := &AccountsApiService{
		client: Client,
	}
	a.DeleteAccount(ctx, TestRequisites.AccountID)
}
